/*
 * Copyright (C) 1997-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file client_common.c common functions of para_client and para_audiod */

#include "para.h"
#include "list.h"
#include "sched.h"
#include "client.cmdline.h"
#include "crypt.h"
#include "rc4.h"
#include "net.h"
#include "fd.h"
#include "string.h"
#include "client.cmdline.h"
#include "client.h"
#include "error.h"

void rc4_send(unsigned long len, const unsigned char *indata,
		unsigned char *outdata, void *private_data)
{
	struct private_client_data *pcd = private_data;
	RC4(&pcd->rc4_send_key, len, indata, outdata);
}

void rc4_recv(unsigned long len, const unsigned char *indata,
		unsigned char *outdata, void *private_data)
{
	struct private_client_data *pcd = private_data;
	RC4(&pcd->rc4_recv_key, len, indata, outdata);
}


void client_close(struct private_client_data *pcd)
{
	if (pcd)
		return;
	if (pcd->fd >= 0) {
		disable_crypt(pcd->fd);
		close(pcd->fd);
	}
	free(pcd->user);
	free(pcd->config_file);
	free(pcd->key_file);
	free(pcd);
}

int client_parse_config(int argc, char *argv[],
		struct private_client_data **pcd_ptr)
{
	char *home = para_homedir();
	struct stat statbuf;
	int ret;
	struct private_client_data *pcd =
		para_calloc(sizeof(struct private_client_data));

	pcd->fd = -1;
	client_cmdline_parser(argc, argv, &pcd->conf);
	ret = - E_CLIENT_SYNTAX;
	if (!pcd->conf.inputs_num)
		goto out;
	pcd->user = pcd->conf.user_given?
		para_strdup(pcd->conf.user_arg) : para_logname();

	pcd->key_file = pcd->conf.key_file_given?
		para_strdup(pcd->conf.key_file_arg) :
		make_message("%s/.paraslash/key.%s", home, pcd->user);

	pcd->config_file = pcd->conf.config_file_given?
		para_strdup(pcd->conf.config_file_arg) :
		make_message("%s/.paraslash/client.conf", home);
	ret = stat(pcd->config_file, &statbuf);
	if (ret && pcd->conf.config_file_given) {
		ret = -E_NO_CONFIG;
		goto out;
	}
	if (!ret)
		client_cmdline_parser_configfile(pcd->config_file,
			&pcd->conf, 0, 0, 0);
	ret = 1;
	*pcd_ptr = pcd;
	PARA_INFO_LOG("loglevel: %d\n", pcd->conf.loglevel_arg);
	PARA_INFO_LOG("config_file: %s\n", pcd->config_file);
	PARA_INFO_LOG("key_file: %s\n", pcd->key_file);
	PARA_NOTICE_LOG("connecting %s:%d\n", pcd->conf.hostname_arg,
		pcd->conf.server_port_arg);
out:
	free(home);
	if (ret < 0)
		client_close(pcd);
	return ret;
}

void client_pre_select(struct sched *s, struct task *t)
{
	struct private_client_data *pcd = t->private_data;

	t->ret = 1;
	pcd->check_r = 0;
	pcd->check_w = 0;
	if (pcd->fd < 0)
		return;
	switch (pcd->status) {
	case CL_CONNECTED:
	case CL_SENT_AUTH:
	case CL_SENT_CH_RESPONSE:
	case CL_SENT_COMMAND:
		para_fd_set(pcd->fd, &s->rfds, &s->max_fileno);
		pcd->check_r = 1;
		return;

	case CL_RECEIVED_WELCOME:
	case CL_RECEIVED_CHALLENGE:
	case CL_RECEIVED_PROCEED:
		para_fd_set(pcd->fd, &s->wfds, &s->max_fileno);
		pcd->check_w = 1;
		return;

	case CL_RECEIVING:
		if (pcd->loaded < CLIENT_BUFSIZE - 1) {
			para_fd_set(pcd->fd, &s->rfds, &s->max_fileno);
			pcd->check_r = 1;
		}
		return;
	case CL_SENDING:
		if (*pcd->in_loaded) {
			PARA_INFO_LOG("loaded: %zd\n", *pcd->in_loaded);
			para_fd_set(pcd->fd, &s->wfds, &s->max_fileno);
			pcd->check_w = 1;
		} else {
			if (*pcd->in_eof) {
				t->ret = -E_INPUT_EOF;
				s->timeout.tv_sec = 0;
				s->timeout.tv_usec = 1;
			}
		}
		return;
	}
}

static ssize_t client_recv_buffer(struct private_client_data *pcd)
{
	ssize_t ret = recv_buffer(pcd->fd, pcd->buf + pcd->loaded,
		CLIENT_BUFSIZE - pcd->loaded);
	if (!ret)
		return -E_SERVER_EOF;
	if (ret > 0)
		pcd->loaded += ret;
	return ret;

}

void client_post_select(struct sched *s, struct task *t)
{
	struct private_client_data *pcd = t->private_data;

//	PARA_INFO_LOG("status %d\n", pcd->status);
	t->ret = 1;
	if (pcd->fd < 0)
		return;
	if (!pcd->check_r && !pcd->check_w)
		return;
	if (pcd->check_r && !FD_ISSET(pcd->fd, &s->rfds))
		return;
	if (pcd->check_w && !FD_ISSET(pcd->fd, &s->wfds))
		return;
	switch (pcd->status) {
	case CL_CONNECTED: /* receive welcome message */
		t->ret = client_recv_buffer(pcd);
		if (t->ret > 0)
			pcd->status = CL_RECEIVED_WELCOME;
		return;
	case CL_RECEIVED_WELCOME: /* send auth command */
		sprintf(pcd->buf, "auth %s%s", pcd->conf.plain_given?
			"" : "rc4 ", pcd->user);
		PARA_INFO_LOG("--> %s\n", pcd->buf);
		t->ret = send_buffer(pcd->fd, pcd->buf);
		if (t->ret >= 0)
			pcd->status = CL_SENT_AUTH;
		return;
	case CL_SENT_AUTH: /* receive challenge number */
		pcd->loaded = 0;
		t->ret = client_recv_buffer(pcd);
		if (t->ret < 0)
			return;
		if (t->ret != 64) {
			t->ret = -E_INVALID_CHALLENGE;
			PARA_ERROR_LOG("received the following: %s\n", pcd->buf);
			return;
		}
		PARA_INFO_LOG("%s", "<-- [challenge]\n");
		/* decrypt challenge number */
		t->ret = para_decrypt_challenge(pcd->key_file, &pcd->challenge_nr,
			(unsigned char *) pcd->buf, 64);
		if (t->ret > 0)
			pcd->status = CL_RECEIVED_CHALLENGE;
		return;
	case CL_RECEIVED_CHALLENGE: /* send decrypted challenge */
		PARA_INFO_LOG("--> %lu\n", pcd->challenge_nr);
		t->ret = send_va_buffer(pcd->fd, "%s%lu", CHALLENGE_RESPONSE_MSG,
			pcd->challenge_nr);
		if (t->ret > 0)
			pcd->status = CL_SENT_CH_RESPONSE;
		return;
	case CL_SENT_CH_RESPONSE: /* read server response */
		{
		size_t bytes_received;
		unsigned char rc4_buf[2 * RC4_KEY_LEN] = "";
		pcd->loaded = 0;
		t->ret = client_recv_buffer(pcd);
		if (t->ret < 0)
			return;
		bytes_received = t->ret;
		PARA_DEBUG_LOG("++++ server info ++++\n%s\n++++ end of server "
			"info ++++\n", pcd->buf);
		/* check if server has sent "Proceed" message */
		t->ret = -E_CLIENT_AUTH;
		if (!strstr(pcd->buf, PROCEED_MSG))
			return;
		t->ret = 1;
		pcd->status = CL_RECEIVED_PROCEED;
		if (bytes_received < PROCEED_MSG_LEN + 32)
			return;
		PARA_INFO_LOG("%s", "decrypting session key\n");
		t->ret = para_decrypt_buffer(pcd->key_file, rc4_buf,
			(unsigned char *)pcd->buf + PROCEED_MSG_LEN + 1,
			bytes_received - PROCEED_MSG_LEN - 1);
		if (t->ret < 0)
			return;
		RC4_set_key(&pcd->rc4_send_key, RC4_KEY_LEN, rc4_buf);
		RC4_set_key(&pcd->rc4_recv_key, RC4_KEY_LEN, rc4_buf + RC4_KEY_LEN);
		enable_crypt(pcd->fd, rc4_recv, rc4_send, pcd);
		}
	case CL_RECEIVED_PROCEED: /* concat args and send command */
		{
		int i;
		char *command = NULL;
		for (i = 0; i < pcd->conf.inputs_num; i++) {
			char *tmp = command;
			command = make_message("%s\n%s", command?
				command : "", pcd->conf.inputs[i]);
			free(tmp);
		}
		command = para_strcat(command, EOC_MSG "\n");
		PARA_DEBUG_LOG("--> %s\n", command);
		t->ret = send_buffer(pcd->fd, command);
		free(command);
		if (t->ret > 0)
			pcd->status = CL_SENT_COMMAND;
		return;
		}
	case CL_SENT_COMMAND:
		pcd->loaded = 0;
		t->ret = client_recv_buffer(pcd);
		if (t->ret < 0)
			return;
		t->ret = -E_HANDSHAKE_COMPLETE;
		if (strstr(pcd->buf, AWAITING_DATA_MSG))
			pcd->status = CL_SENDING;
		else
			pcd->status = CL_RECEIVING;
		return;
	case CL_SENDING: /* FIXME: might block */
		PARA_INFO_LOG("loaded: %zd\n", *pcd->in_loaded);
		t->ret = send_bin_buffer(pcd->fd, pcd->inbuf, *pcd->in_loaded);
		if (t->ret < 0)
			return;
		*pcd->in_loaded = 0; /* FIXME: short writes */
		return;
	case CL_RECEIVING:
		t->ret = client_recv_buffer(pcd);
		return;
	}

}

int client_open(struct private_client_data *pcd)
{
	int ret;
	struct hostent *he;
	struct sockaddr_in their_addr;

	pcd->fd = -1;
	ret = get_host_info(pcd->conf.hostname_arg, &he);
	if (ret < 0)
		goto err_out;
	/* get new socket */
	ret = get_socket();
	if (ret < 0)
		goto err_out;
	pcd->fd = ret;
	/* init their_addr */
	init_sockaddr(&their_addr, pcd->conf.server_port_arg, he);
	ret = para_connect(pcd->fd, &their_addr);
	if (ret < 0)
		goto err_out;
	pcd->status = CL_CONNECTED;
	ret = mark_fd_nonblock(pcd->fd);
	if (ret < 0)
		goto err_out;
	pcd->task.pre_select = client_pre_select;
	pcd->task.post_select = client_post_select;
	pcd->task.private_data = pcd;
	sprintf(pcd->task.status, "client");
	register_task(&pcd->task);
	return 1;
err_out:
	if (pcd->fd >= 0)
		close(pcd->fd);
	return ret;
}
